import urllib.parse

from tango import DevState, GreenMode
from tango.server import Device, attribute, command, device_property
from connio import connection_for_url

from ... import dcp3000


ATTR_MAP = {
    "version": lambda dcp: dcp.version(),
    "errors": lambda dcp: dcp.errors(),
    "pressure": lambda dcp: dcp.pressure(),
    "transducer_pressures": lambda dcp: dcp.transducer_pressures(),
    "transducer_pressure_1": lambda dcp: dcp.transducer_pressure(1),
    "transducer_pressure_2": lambda dcp: dcp.transducer_pressure(2),
    "transducer_pressure_3": lambda dcp: dcp.transducer_pressure(3),
    "transducer_pressure_4": lambda dcp: dcp.transducer_pressure(4),
    "on_setpoint_1": lambda dcp: dcp.on_setpoint(1),
    "on_setpoint_2": lambda dcp: dcp.on_setpoint(2),
    "on_setpoint_3": lambda dcp: dcp.on_setpoint(3),
    "on_setpoint_4": lambda dcp: dcp.on_setpoint(4),
    "off_setpoint_1": lambda dcp: dcp.off_setpoint(1),
    "off_setpoint_2": lambda dcp: dcp.off_setpoint(2),
    "off_setpoint_3": lambda dcp: dcp.off_setpoint(3),
    "off_setpoint_4": lambda dcp: dcp.off_setpoint(4),
}


class DCP3000(Device):

    green_mode = GreenMode.Asyncio

    url = device_property(dtype=str)
    baudrate = device_property(dtype=int, default_value=9600)
    bytesize = device_property(dtype=int, default_value=8)
    parity = device_property(dtype=str, default_value="N")

    dcp = None

    def url_to_connection_args(self):
        url = self.url
        res = urllib.parse.urlparse(url)
        kwargs = dict(concurrency="async")
        if res.scheme in {"serial", "rfc2217"}:
            kwargs.update(
                dict(baudrate=self.baudrate, bytesize=self.bytesize, parity=self.parity)
            )
        return url, kwargs

    async def init_device(self):
        await super().init_device()
        url, kwargs = self.url_to_connection_args()
        conn = connection_for_url(url, **kwargs)
        self.dcp = dcp3000.DCP3000(conn)
        self.last_values = {}

    async def read_attr_hardware(self, indexes):
        multi_attr = self.get_device_attr()
        names = [
            multi_attr.get_attr_by_ind(index).get_name().lower() for index in indexes
        ]
        funcs = (ATTR_MAP[name] for name in names)
        values = [await func(self.dcp) for func in funcs]
        self.last_values = dict(zip(names, values))

    async def dev_state(self):
        state = DevState.ON
        try:
            errors = await self.dcp.errors()
            if errors:
                state = DevState.ALARM
        except:
            state = DevState.FAULT
        return state

    async def dev_status(self):
        self.__status = "Ready!"
        try:
            errors = await self.dcp.errors()
            if errors:
                self.__status = "Hardware error(s):\n" + "\n".join(errors)
        except Exception as error:
            self.__status = "Communication error:\n{!r}".format(error)
        return self.__status

    @attribute(unit="mbar")
    def pressure(self):
        return self.last_values["pressure"]

    @attribute(unit="mbar", dtype=[float], max_dim_x=8)
    def transducer_pressures(self):
        return self.last_values["transducer_pressures"]

    @attribute(unit="mbar")
    def transducer_pressure_1(self):
        return self.last_values["transducer_pressure_1"]

    @attribute(unit="mbar")
    def transducer_pressure_2(self):
        return self.last_values["transducer_pressure_2"]

    @attribute(unit="mbar")
    def transducer_pressure_3(self):
        return self.last_values["transducer_pressure_3"]

    @attribute(unit="mbar")
    def transducer_pressure_4(self):
        return self.last_values["transducer_pressure_4"]

    @attribute(dtype=[str])
    def errors(self):
        return self.last_values["errors"]

    @attribute(dtype=str)
    def version(self):
        return self.last_values["version"]

    @attribute()
    def on_setpoint_1(self):
        return self.last_values["on_setpoint_1"]

    @on_setpoint_1.setter
    def on_setpoint_1(self, value):
        return self.dcp.on_setpoint(1, value)

    @attribute()
    def on_setpoint_2(self):
        return self.last_values["on_setpoint_2"]

    @on_setpoint_2.setter
    def on_setpoint_2(self, value):
        return self.dcp.on_setpoint(2, value)

    @attribute()
    def on_setpoint_3(self):
        return self.last_values["on_setpoint_3"]

    @on_setpoint_3.setter
    def on_setpoint_3(self, value):
        return self.dcp.on_setpoint(3, value)

    @attribute()
    def on_setpoint_4(self):
        return self.last_values["on_setpoint_4"]

    @on_setpoint_4.setter
    def on_setpoint_4(self, value):
        return self.dcp.on_setpoint(4, value)

    @attribute()
    def off_setpoint_1(self):
        return self.last_values["off_setpoint_1"]

    @off_setpoint_1.setter
    def off_setpoint_1(self, value):
        return self.dcp.off_setpoint(1, value)

    @attribute()
    def off_setpoint_2(self):
        return self.last_values["off_setpoint_2"]

    @off_setpoint_2.setter
    def off_setpoint_2(self, value):
        return self.dcp.off_setpoint(2, value)

    @attribute()
    def off_setpoint_3(self):
        return self.last_values["off_setpoint_3"]

    @off_setpoint_3.setter
    def off_setpoint_3(self, value):
        return self.dcp.off_setpoint(3, value)

    @attribute()
    def off_setpoint_4(self):
        return self.last_values["off_setpoint_4"]

    @off_setpoint_4.setter
    def off_setpoint_4(self, value):
        return self.dcp.off_setpoint(4, value)

    @command
    def switch_on(self):
        return self.dcp.switch_on()

    @command
    def switch_off(self):
        return self.dcp.switch_off()

    @command
    def open_venting_valve(self):
        return self.dcp.open_venting_value()

    @command
    def close_venting_valve(self):
        return self.dcp.close_venting_valve()

    @command
    def vent(self):
        return self.dcp.vent()


def main():
    DCP3000.run_server()


if __name__ == "__main__":
    main()



from .dcp3000 import DCP3000


def main():
    import sys
    import logging
    from tango import GreenMode
    from tango.server import run

    args = ["Vacuubrand"] + sys.argv[1:]
    fmt = "%(asctime)s %(threadName)s %(levelname)s %(name)s %(message)s"
    logging.basicConfig(level=logging.INFO, format=fmt)
    run((DCP3000,), args=args, green_mode=GreenMode.Asyncio)



# -*- coding: utf-8 -*-
#
# This file is part of the vacuubrand project
#
# Copyright (c) 2020 Tiago Coutinho
# Distributed under the LGPLv3. See LICENSE for more info.

"""
.. code-block:: yaml

    devices:
    - class: DCP3000
      package: vacuubrand.simulator
      transports:
      - type: serial
        url: /tmp/dcp3000-1
"""

import time

from sinstruments.simulator import BaseDevice


DEFAULT = {
    "IN_VER": "DCP 3000  V2.30",
    "IN_PV_S1": "0234 mbar",
    "IN_PV_S2": "0278 mbar",
    "IN_PV_S3": "0213 mbar",
    "IN_PV_S4": "0223 mbar",
    "IN_PV_X": "0362 0234 0278 0213 mbar",
    # English, mbar, no acoustic signal, venting valve, fault indicator, 4 transducers
    "IN_CFG": "1001144",
    "IN_ERR": "00000",
    "IN_SP_1": "00:00",  # time interval for sending
    "IN_SP_2": "99:00",  # time interval for recording
    "IN_SP_11": "111.0 mbar",  # On setpoint channel 1
    "IN_SP_21": "011.5 mbar",  # Off setpoint channel 1
    "IN_SP_12": "222.0 mbar",  # On setpoint channel 1
    "IN_SP_22": "022.5 mbar",  # Off setpoint channel 1
    "IN_SP_13": "333.0 mbar",  # On setpoint channel 1
    "IN_SP_23": "033.5 mbar",  # Off setpoint channel 1
    "IN_SP_14": "444.0 mbar",  # On setpoint channel 1
    "IN_SP_24": "044.5 mbar",  # Off setpoint channel 1
}


def to_hh_ss(text):
    if ":" in text:
        return text
    sec = int(text)
    return "{:02d}:{:02d}".format(sec // 60, sec % 60)


class DCP3000(BaseDevice):
    def __init__(self, name, **opts):
        kwargs = {}
        if "newline" in opts:
            kwargs["newline"] = opts.pop("newline")
        self._config = dict(DEFAULT, **opts)
        self._on = True
        super().__init__(name, **kwargs)

    def handle_message(self, line):
        self._log.debug("request %r", line)
        line = line.decode().strip()
        result = None
        cmd, *args = line.split()
        if cmd == "OUT_SP_1":
            self._config["IN_SP_1"] = to_hh_ss(args[0])
        elif line.startswith("OUT_SP_2"):
            self._config["IN_SP_2"] = to_hh_ss(args[0])
        elif cmd == "OUT_VENT":
            pass
        elif cmd == "OUT_SENSOR":
            pass
        elif cmd == "REMOTE":
            self._on = args[0] == "1"
        elif cmd.startswith("OUT_SP_"):
            self._config["IN_SP_" + cmd[7:9]] = args[0]
        elif line == "IN_PV_1":
            v = "1004.1" if self._on else "0000.0"
            result = v + " mbar"
        else:
            result = self._config.get(line)
        if result is None:
            return
        result = result.encode() + b"\r\n"
        self._log.debug("reply %r", result)
        return result



import enum
import time
import asyncio
import logging
import functools
import threading

from connio import connection_for_url


UNITS = {"0": "mbar", "1": "Torr", "2": "hPa"}


ERRORS = [
    "venting valve fault",
    "overpressure",
    "pressure transducer fault",
    "external fault",
    #    'last serial command incorrect'
]


torr_to_mbar = 1.3332236842105263


def syncer(func):
    async def acall(coro):
        return func(await coro)

    @functools.wraps(func)
    def wrapper(arg):
        return acall(arg) if asyncio.iscoroutine(arg) else func(arg)

    return wrapper


@syncer
def decode_config(text):
    assert len(text) == 7, text
    return dict(
        unit=UNITS[text[1]],
        acoustic_signal=text[2] == "1",
        venting_valve_connected=text[3] == "1",
        fault_indicator_connected=text[4] == "1",
        nb_active_pressure_transducers=int(text[5]),
        nb_pressure_transducers=int(text[6]),
    )


@syncer
def decode_pressure(text):
    value, unit = text.lower().split()
    value = float(value)
    if unit == "torr":
        value *= torr_to_mbar
    return value


@syncer
def decode_pressures(text):
    """decode pressure(s). Always return values in millibar"""
    *values, unit = text.lower().split()
    values = [float(value) for value in values]
    if unit == "torr":
        values = [value * torr_to_mbar for value in values]
    return values


@syncer
def decode_errors(text):
    """return a list of errors"""
    assert len(text) == 5, text
    text = text[:-1]
    return [error for i, error in zip(text, ERRORS) if i == "1"]


@syncer
def decode_interval(text):
    minutes, seconds = [int(v) for v in text[:4].split(":")]
    return minutes * 60 + seconds


def encode(request):
    return request.encode() + b"\n"


def decode(reply):
    return reply.decode().strip()


class BaseProtocol:
    """
    Handles communication protocol
    - latency / back-pressure
    - encode/decode bytes <-> text
    - serializes read calls
    """

    # max 20 commands/s (according to doc) <=> 50 ms (put 60 to be safe side)
    COMMAND_LATENCY = 0.06

    def __init__(self, connection, log=None):
        self.conn = connection
        self._last_command = 0
        self._log = log or logging.getLogger(
            "vacuubrand.dcp3000.{}".format(type(self).__name__)
        )

    def _wait_time(self):
        return self._last_command + self.COMMAND_LATENCY - time.monotonic()


class AIOProtocol(BaseProtocol):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._lock = asyncio.Lock()

    async def _back_pressure(self):
        wait = self._wait_time()
        if wait > 0:
            await asyncio.sleep(wait)

    async def write_readline(self, data):  # aka: query or put_get
        data = encode(data)
        self._log.debug("write: %r", data)
        await self._back_pressure()
        try:
            async with self._lock:
                # TODO: maybe consume garbage in the buffer ?
                reply = await self.conn.write_readline(data)
            self._log.debug("read: %r", reply)
            return decode(reply)
        finally:
            self._last_command = time.monotonic()

    async def write(self, data):
        data = encode(data)
        self._log.debug("write: %r", data)
        await self._back_pressure()
        try:
            await self.conn.write(data)
        finally:
            self._last_command = time.monotonic()


class IOProtocol(BaseProtocol):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._lock = threading.Lock()

    def _back_pressure(self):
        wait = self._wait_time()
        if wait > 0:
            time.sleep(wait)

    def write_readline(self, data):  # aka: query or put_get
        data = encode(data)
        self._log.debug("write: %r", data)
        self._back_pressure()
        try:
            with self._lock:
                # TODO: maybe consume garbage in the buffer ?
                reply = self.conn.write_readline(data)
            self._log.debug("read: %r", reply)
            return decode(reply)
        finally:
            self._last_command = time.monotonic()

    def write(self, data):  # aka: query or put_get
        data = encode(data)
        self._log.debug("write: %r", data)
        self._back_pressure()
        try:
            self.conn.write(data)
        finally:
            self._last_command = time.monotonic()


def Protocol(connection, *args, **kwargs):
    func = connection.write_readline
    klass = AIOProtocol if asyncio.iscoroutinefunction(func) else IOProtocol
    return klass(connection, *args, **kwargs)


def protocol_for_url(url, *args, **kwargs):
    log = kwargs.pop("log", None)
    conn = connection_for_url(url, *args, **kwargs)
    return Protocol(conn, log=log)



import logging

from .protocol import (
    Protocol,
    decode_config,
    decode_pressure,
    decode_pressures,
    decode_interval,
    decode_errors,
)


class DCP3000:
    def __init__(self, connection):
        self._log = logging.getLogger("xia_pfcu.{}".format(type(self).__name__))
        self.protocol = Protocol(connection, log=self._log)

    def write_readline(self, request):
        return self.protocol.write_readline(request)

    def write(self, request):
        return self.protocol.write(request)

    def config(self):
        reply = self.write_readline("IN_CFG")
        return decode_config(reply)

    def pressure(self):
        reply = self.write_readline("IN_PV_1")
        return decode_pressure(reply)

    def transducer_pressure(self, channel=1):
        reply = self.write_readline("IN_PV_S{}".format(channel))
        return decode_pressure(reply)

    def transducer_pressures(self):
        reply = self.write_readline("IN_PV_X")
        return decode_pressures(reply)

    def event_time_interval(self, value: int = None):
        if value is not None:
            return self.write("OUT_SP_1 {}".format(value))
        reply = self.write_readline("IN_SP_1")
        return decode_interval(reply)

    def recording_time_interval(self, value: int = None):
        if value is not None:
            self.write("OUT_SP_2 {}".format(value))
        reply = self.write_readline("IN_SP_2")
        return decode_interval(reply)

    def _set_point(self, channel: int, on_off: bool, value: (float, "mbar") = None):
        assert channel in {1, 2, 3, 4}
        on_off = 1 if on_off else 2
        if value is not None:
            return self.write("OUT_SP_{}{} {} mbar".format(on_off, channel, value))
        reply = self.write_readline("IN_SP_{}{}".format(on_off, channel))
        return decode_pressure(reply)

    def on_setpoint(self, channel: int, value: (float, "mbar") = None):
        return self._set_point(channel, True, value=value)

    def off_setpoint(self, channel: int, value: (float, "mbar") = None):
        return self._set_point(channel, False, value=value)

    def errors(self):
        reply = self.write_readline("IN_ERR")
        return decode_errors(reply)

    def version(self):
        return self.write_readline("IN_VER")

    def switch_on(self):
        return self.write("REMOTE 1")

    def switch_off(self):
        return self.write("REMOTE -")

    def close_venting_valve(self):
        return self.write("OUT_VENT 0")

    def open_venting_value(self):
        return self.write("OUT_VENT 1")

    def vent(self):
        """venting until atmospheric pressure"""
        return self.write("OUT_VENT 2")



# -*- coding: utf-8 -*-

"""The setup script."""

import sys
from setuptools import setup, find_packages

requirements = ["connio"]

with open("README.md") as f:
    description = f.read()


setup(
    name="vacuubrand",
    author="Tiago Coutinho",
    author_email="tcoutinho@cells.es",
    version="2.1.2",
    description="vacuubrand library",
    long_description=description,
    long_description_content_type="text/markdown",
    entry_points={
        "console_scripts": [
            "Vacuubrand = vacuubrand.tango.server:main [tango]",
        ]
    },
    extras_require={
        "tango": ["pytango>=9"],
        "simulator": ["sinstruments>=1"]
    },
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8"
    ],
    install_requires=requirements,
    license="LGPLv3",
    include_package_data=True,
    keywords="vacuubrand, library, tango",
    packages=find_packages(),
    url="https://github.com/tiagocoutinho/vacuubrand",
    project_urls={
        "Documentation": "https://github.com/tiagocoutinho/vacuubrand",
        "Source": "https://github.com/tiagocoutinho/vacuubrand"
    }
)
